# Needed so that mac wheels properly delocate:
#  https://github.com/matthew-brett/delocate/issues/22

from libsonata._libsonata import (EdgePopulation,
                                  EdgeStorage,
                                  ElementDataFrame,
                                  ElementReportPopulation,
                                  ElementReportReader,
                                  NodePopulation,
                                  NodeStorage,
                                  Selection,
                                  SomaDataFrame,
                                  SomaReportPopulation,
                                  SomaReportReader,
                                  SonataError,
                                  SpikePopulation,
                                  SpikeReader,
                                  version,
                                  )



import os
import unittest

import numpy as np

from libsonata import EdgeStorage, NodeStorage, Selection, SonataError
from libsonata import SpikeReader, SpikePopulation, SomaReportReader, SomaReportPopulation, ElementReportReader, ElementReportPopulation


PATH = os.path.dirname(os.path.realpath(__file__))
PATH = os.path.join(PATH, '../../tests/data')


class TestSelection(unittest.TestCase):
    def test_basic(self):
        ranges = [(3, 5), (0, 3)]
        selection = Selection(ranges)
        self.assertTrue(selection)
        self.assertEqual(selection.ranges, ranges)
        self.assertEqual(selection.flat_size, 5)
        self.assertEqual(selection.flatten().tolist(), [3, 4, 0, 1, 2])

    def test_from_values(self):
        values = [
            [1, 3, 4, 1],
            [0, 0, 0, 0],
        ]
        expected = [(1, 2), (3, 5), (1, 2)]
        self.assertEqual(Selection(values[0]).ranges, expected)
        self.assertEqual(Selection(np.array(values, dtype=np.uint64, order='C')[0]).ranges, expected)
        self.assertEqual(Selection(np.array(values, dtype=np.uint32, order='C')[0]).ranges, expected)
        self.assertEqual(Selection(np.array(values, dtype=np.uint64, order='F')[0]).ranges, expected)

    def test_from_values_empty(self):
        self.assertFalse(Selection([]))
        self.assertFalse(Selection(np.array([], dtype=np.uint64)))

    def test_comparison(self):
        empty = Selection([])
        range_selection = Selection(((0, 2), (3, 4)))
        values_selection = Selection([1, 3, 4, 1])

        self.assertEqual(empty, empty)
        self.assertNotEqual(empty, range_selection)
        self.assertNotEqual(empty, values_selection)

        self.assertEqual(range_selection, range_selection)
        self.assertNotEqual(range_selection, values_selection)

        self.assertEqual(values_selection, values_selection)

        values_selection1 = Selection([1, 3, 4, 1])
        self.assertEqual(values_selection, values_selection1)

    def test_union_intersection(self):
        empty = Selection([])
        self.assertEqual(empty, empty & empty)
        self.assertEqual(empty, empty | empty)

        even = Selection(list(range(0, 10, 2)))
        odd = Selection(list(range(1, 10, 2)))
        self.assertEqual(empty, empty & even)
        self.assertEqual(even, empty | even)
        self.assertEqual(empty, odd & even)
        self.assertEqual(Selection(list(range(10))), odd | even)


class TestNodePopulation(unittest.TestCase):
    def setUp(self):
        path = os.path.join(PATH, 'nodes1.h5')
        self.test_obj = NodeStorage(path).open_population('nodes-A')

    def test_name(self):
        self.assertEqual(self.test_obj.name, "nodes-A")

    def test_size(self):
        self.assertEqual(self.test_obj.size, 6)
        self.assertEqual(len(self.test_obj), 6)

    def test_attribute_names(self):
        self.assertEqual(
            self.test_obj.attribute_names,
            {
                'attr-X', 'attr-Y', 'attr-Z',
                'A-int8', 'A-int16', 'A-int32', 'A-int64',
                'A-uint8', 'A-uint16', 'A-uint32', 'A-uint64',
                'A-float', 'A-double',
                'A-string',
                'A-enum',
                'E-mapping-good',
                'E-mapping-bad'
            }
        )

    def test_get_attribute(self):
        self.assertEqual(self.test_obj.get_attribute('attr-X', 0), 11.)
        self.assertEqual(self.test_obj.get_attribute('attr-X', Selection([0, 5])).tolist(), [11., 16.])

        # different dtypes
        self.assertEqual(self.test_obj.get_attribute('attr-Y', 0), 21)
        self.assertEqual(self.test_obj.get_attribute('attr-Z', 0), 'aa')

        # default value
        self.assertEqual(self.test_obj.get_attribute('attr-X', Selection([0, 5]), 42.).tolist(), [11., 16.])

        self.assertRaises(SonataError, self.test_obj.get_attribute, 'no-such-attribute', 0)

    def test_get_dynamics_attribute(self):
        self.assertEqual(self.test_obj.get_dynamics_attribute('dparam-X', 0), 1011.)
        self.assertEqual(self.test_obj.get_dynamics_attribute('dparam-X', Selection([0, 5])).tolist(), [1011., 1016.])

        # different dtypes
        self.assertEqual(self.test_obj.get_dynamics_attribute('dparam-Y', 0), 1021)
        self.assertEqual(self.test_obj.get_dynamics_attribute('dparam-Z', 0), 'd-aa')

        # default value
        self.assertEqual(self.test_obj.get_dynamics_attribute('dparam-X', Selection([0, 5]), 42.).tolist(), [1011., 1016.])

        self.assertRaises(SonataError, self.test_obj.get_dynamics_attribute, 'no-such-attribute', 0)

    def test_enumeration_names(self):
        self.assertEqual(
            self.test_obj.enumeration_names,
            {
                'E-mapping-good',
                'E-mapping-bad'
            }
        )

    def test_enumeration_values(self):
        self.assertEqual(
            self.test_obj.get_attribute(
                "E-mapping-good",
                Selection([(0, 1), (2, 3)])
            ).tolist(),
            ["C", "C"]
        )

        self.assertEqual(
            self.test_obj.get_attribute(
                "E-mapping-good",
                Selection([(0, 1), (2, 3)])
            ).tolist(),
            ["C", "C"]
        )

        self.assertEqual(
            self.test_obj.get_enumeration(
                "E-mapping-good",
                Selection([(0, 1), (2, 3)])
            ).tolist(),
            [2, 2]
        )

        self.assertEqual(
            self.test_obj.enumeration_values("E-mapping-good"),
            ["A", "B", "C"]
        )

        self.assertRaises(
            SonataError,
            self.test_obj.get_attribute,
            "E-mapping-bad",
            1
        )

        self.assertRaises(
            SonataError,
            self.test_obj.get_enumeration,
            "attr-X",
            0
        )

    def test_select_all(self):
        self.assertEqual(self.test_obj.select_all().flat_size, 6)

    def test_match_values(self):
        # string
        self.assertEqual(self.test_obj.match_values("attr-Z", "bb").flatten().tolist(),
                         [1])

        # enum
        self.assertEqual(self.test_obj.match_values("E-mapping-good", "C").flatten().tolist(),
                         [0, 2, 4, 5])

        # int
        self.assertEqual(self.test_obj.match_values("attr-Y", 23).flatten().tolist(),
                         [2, ])

        # float
        self.assertRaises(TypeError, self.test_obj.match_values, "attr-Y", 23.)


class TestEdgePopulation(unittest.TestCase):
    def setUp(self):
        path = os.path.join(PATH, 'edges1.h5')
        self.test_obj = EdgeStorage(path).open_population('edges-AB')

    def test_source(self):
        self.assertEqual(self.test_obj.source, 'nodes-A')

    def test_target(self):
        self.assertEqual(self.test_obj.target, 'nodes-B')

    def test_source_nodes(self):
        self.assertEqual(self.test_obj.source_node(1), 1)
        self.assertEqual(self.test_obj.source_nodes(Selection([0, 1, 2, 4])).tolist(), [1, 1, 2, 3])

    def test_target_nodes(self):
        self.assertEqual(self.test_obj.target_node(1), 2)
        self.assertEqual(self.test_obj.target_nodes(Selection([0, 1, 2, 4])).tolist(), [1, 2, 1, 0])

    def test_afferent_edges(self):
        self.assertEqual(self.test_obj.afferent_edges([1, 2]).ranges, [(0, 4), (5, 6)])
        self.assertEqual(self.test_obj.afferent_edges(1).ranges, [(0, 1), (2, 4)])

    def test_efferent_edges(self):
        self.assertEqual(self.test_obj.efferent_edges([1, 2]).ranges, [(0, 4)])
        self.assertEqual(self.test_obj.efferent_edges(0).ranges, [])

    def test_connecting_edges(self):
        self.assertEqual(self.test_obj.connecting_edges([1, 2], [1, 2]).ranges, [(0, 4)])
        self.assertEqual(self.test_obj.connecting_edges(1, 1).ranges, [(0, 1)])

    def test_select_all(self):
        self.assertEqual(self.test_obj.select_all().flat_size, 6)


class TestSpikePopulation(unittest.TestCase):
    def setUp(self):
        path = os.path.join(PATH, "spikes.h5")
        self.test_obj = SpikeReader(path)

    def test_get_all_populations(self):
        self.assertEqual(self.test_obj.get_population_names(), ['All', 'empty', 'spikes1', 'spikes2'])

    def test_get_population(self):
        self.assertTrue(isinstance(self.test_obj['spikes1'], SpikePopulation))

    def test_get_inexistant_population(self):
        self.assertRaises(RuntimeError, self.test_obj.__getitem__, 'foobar')

    def test_get_spikes_from_population(self):
        self.assertEqual(self.test_obj['All'].get(), [(5, 0.1), (2, 0.2), (3, 0.3), (2, 0.7), (3, 1.3)])
        self.assertEqual(self.test_obj['All'].get(tstart=0.2, tstop=1.0), [(2, 0.2), (3, 0.3), (2, 0.7)])
        self.assertEqual(self.test_obj['spikes2'].get(tstart=0.2, tstop=1.0), [(3, 0.3), (2, 0.2), (2, 0.7)])
        self.assertEqual(self.test_obj['spikes1'].get((3,)), [(3, 0.3), (3, 1.3)])
        self.assertEqual(self.test_obj['spikes2'].get((3,)), [(3, 0.3), (3, 1.3)])
        self.assertEqual(self.test_obj['spikes2'].get((10,)), [])
        self.assertEqual(self.test_obj['spikes2'].get((2,), 0., 0.5), [(2, 0.2)])
        self.assertEqual(self.test_obj['spikes1'].get((2, 5)), [(2, 0.2), (2, 0.7), (5, 0.1)])
        self.assertEqual(self.test_obj['spikes2'].get((2, 5)), [(5, 0.1), (2, 0.2), (2, 0.7)])
        self.assertEqual(self.test_obj['All'].sorting, "by_time")
        self.assertEqual(self.test_obj['spikes1'].sorting, "by_id")
        self.assertEqual(self.test_obj['spikes2'].sorting, "none")
        self.assertEqual(self.test_obj['empty'].get(), [])

        self.assertEqual(len(self.test_obj['All'].get(node_ids=[])), 0)

class TestSomaReportPopulation(unittest.TestCase):
    def setUp(self):
        path = os.path.join(PATH, "somas.h5")
        self.test_obj = SomaReportReader(path)

    def test_get_all_population(self):
        self.assertEqual(self.test_obj.get_population_names(), ['All', 'soma1', 'soma2'])

    def test_get_population(self):
        self.assertTrue(isinstance(self.test_obj['All'], SomaReportPopulation))

    def test_get_inexistant_population(self):
        self.assertRaises(RuntimeError, self.test_obj.__getitem__, 'foobar')

    def test_get_reports_from_population(self):
        self.assertEqual(self.test_obj['All'].times, (0., 1., 0.1))
        self.assertEqual(self.test_obj['All'].time_units, 'ms')
        self.assertEqual(self.test_obj['All'].data_units, 'mV')
        self.assertTrue(self.test_obj['All'].sorted)
        self.assertEqual(len(self.test_obj['All'].get().ids), 20)  # Number of nodes
        self.assertEqual(len(self.test_obj['All'].get().times), 10)  # number of times
        self.assertEqual(len(self.test_obj['All'].get().data), 10)  # should be the same
        sel = self.test_obj['All'].get(node_ids=[13, 14], tstart=0.8, tstop=1.0)
        self.assertEqual(len(sel.times), 2)  # Number of timestamp (0.8 and 0.9)
        self.assertEqual(list(sel.ids), [13, 14])

class TestElementReportPopulation(unittest.TestCase):
    def setUp(self):
        path = os.path.join(PATH, "elements.h5")
        self.test_obj = ElementReportReader(path)

    def test_get_all_population(self):
        self.assertEqual(self.test_obj.get_population_names(), ['All', 'element1', 'element42'])

    def test_get_population(self):
        self.assertTrue(isinstance(self.test_obj['All'], ElementReportPopulation))

    def test_get_inexistant_population(self):
        self.assertRaises(RuntimeError, self.test_obj.__getitem__, 'foobar')

    def test_get_node_ids(self):
        self.assertEqual(self.test_obj['All'].get_node_ids(), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20])

    def test_get_reports_from_population(self):
        self.assertEqual(self.test_obj['All'].times, (0., 4., 0.2))
        # check following calls succeed (no memory destroyed)
        self.assertEqual(self.test_obj['All'].times, (0., 4., 0.2))

        self.assertEqual(self.test_obj['All'].time_units, 'ms')
        self.assertEqual(self.test_obj['All'].data_units, 'mV')
        self.assertTrue(self.test_obj['All'].sorted)
        self.assertEqual(len(self.test_obj['All'].get().data), 20)  # Number of times in this range
        self.assertEqual(len(self.test_obj['All'].get().times), 20)  # Should be the same
        self.assertEqual(len(self.test_obj['All'].get().ids), 100)
        sel = self.test_obj['All'].get(node_ids=[13, 14], tstart=0.8, tstop=1.2)
        keys = list(sel.ids)
        keys.sort()
        self.assertEqual(keys, [(13, 30), (13, 30), (13, 31), (13, 31), (13, 32), (14, 32), (14, 33), (14, 33), (14, 34), (14, 34)])

        self.assertEqual(len(self.test_obj['All'].get(node_ids=[]).data), 0)
        self.assertEqual(len(self.test_obj['All'].get(node_ids=[]).times), 0)
        self.assertEqual(len(self.test_obj['All'].get(node_ids=[]).ids), 0)

        self.assertEqual(len(sel.times), 3)  # Number of timestamp (0.8, 1.0 and 1.2)
        with self.assertRaises(SonataError):
            self.test_obj['All'].get(tstart=5.)  # tstart out of range
        np.testing.assert_allclose(self.test_obj['All'].get(node_ids=[1, 2], tstart=3., tstop=3.).data[0], [150.0, 150.1, 150.2, 150.3, 150.4, 150.5, 150.6, 150.7, 150.8, 150.9]) # tstart should be <= tstop
        # check following calls succeed (no memory destroyed)
        np.testing.assert_allclose(self.test_obj['All'].get(node_ids=[1, 2], tstart=3., tstop=3.).data[0], [150.0, 150.1, 150.2, 150.3, 150.4, 150.5, 150.6, 150.7, 150.8, 150.9])
        np.testing.assert_allclose(self.test_obj['All'].get(node_ids=[3, 4], tstart=0.2, tstop=0.4).data[0], [11.0, 11.1, 11.2, 11.3, 11.4, 11.5, 11.6, 11.7, 11.8, 11.9], 1e-6, 0)

if __name__ == '__main__':
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

import os
from pathlib import Path
import pkg_resources
import re
import shutil
import subprocess

from jinja2 import Template

project = "libsonata"
author = "Blue Brain Project, EPFL"
version = pkg_resources.get_distribution("libsonata").version
release = version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.napoleon"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

master_doc = "index"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx-bluebrain-theme"
html_title = "libsonata"
html_show_sourcelink = False
html_extra_path = ["doxygen"]
html_theme_options = {
    "repo_url": "https://github.com/BlueBrain/libsonata/",
    "repo_name": "BlueBrain/libsonata",
}

autodoc_default_options = {"members": True, "imported-members": True}
autodoc_docstring_signature = True

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']


def doxygen_build(app):
    """Build the doxygen output for inclusion in the sphinx docs."""
    subprocess.call(["doxygen", "Doxyfile"], cwd=app.srcdir)


def render_doxygen_pages(app):
    """Render the templated output."""
    NAMESPACE_SEP = "_1_1"
    NAMESPACE_ROOT = "bbp"

    root_dir = Path(app.confdir)
    template_dir = root_dir / "_templates"
    output_path = root_dir / "cpp"
    doxygen_path = root_dir / "doxygen" / "cpp"

    if output_path.exists():
        shutil.rmtree(output_path)

    output_path.mkdir()

    template_file = template_dir / "doxygen_page.rst_t"
    page_template = Template(template_file.read_text())

    for f in doxygen_path.glob("*.html"):
        if f.name.endswith("-members.html"):
            continue

        if NAMESPACE_SEP not in f.name or NAMESPACE_ROOT not in f.name:
            continue

        type_, name = f.name.split(NAMESPACE_ROOT, maxsplit=1)
        name = NAMESPACE_ROOT + name

        name = "".join(
            s.capitalize() for s in name.replace(NAMESPACE_SEP, "::").split("_")
        ).rstrip(".html")
        name = name[0].lower() + name[1:]
        name = name.replace("bbp::sonata::", "")

        filename = "doxygen_" + name
        context = {"title": name, "file": f.name}

        output_file = output_path / (filename + ".rst")
        output_file.write_text(page_template.render(**context))


NUMBERED_LINE_RE = re.compile(r"^([1-9]?\d*)(\. )(.*)$")
SELF_ARG_RE = re.compile(r"(self[^),]*(?:, )?)")
FUNCTION_PARTS_RE = re.compile(r"(.*)\((.*)\)")


def _format_signature(match):
    """Apply the correct RST formatting to the signature."""

    sig = "``" + match.group(1) + "``\ ("

    if match.group(2):
        sig += "*" + match.group(2) + "*"

    sig += "\ )"

    return sig


def _process_pybind_docstrings(app, what, name, obj, options, lines):
    """Pre-process the docstrings generated by pybind."""

    # don't show members as being members of libsonata._libsonata.*
    lines[:] = [l.replace("._libsonata", "") for l in lines]

    # pybind uses "Throws" for some reason, change to "Raises"
    if lines:
        lines[:] = ["Raises:" if l == "Throws:" else l for l in lines]

    # only modify pybind overloaded functions
    if lines and lines[0] == "Overloaded function.":
        for index, line in enumerate(lines[1:], start=1):
            # each overload is numbered so check we have a numbered line
            match = NUMBERED_LINE_RE.match(line)

            # we need to indent lines inside overloads since pybind does not
            if match is None:
                if line:
                    lines[index] = "   " + line
                continue

            # remove self arg
            sig = SELF_ARG_RE.sub("", match.group(3))

            # apply formatting
            sig = FUNCTION_PARTS_RE.sub(_format_signature, sig)

            # replace with arrow unicode charater
            sig = sig.replace("->", "\u2192")

            # rebuild the numbered line
            lines[index] = "#" + match.group(2) + sig


def _process_pybind_signature(
    app, what, name, obj, options, signature, return_annotation
):
    """Pre-process signatures to remove the self argument."""

    if signature is None:
        return

    # don't show members as being members of libsonata._libsonata.*
    signature = signature.replace("._libsonata", "")

    if return_annotation:
        return_annotation = return_annotation.replace("._libsonata", "")

    sig = SELF_ARG_RE.sub("", signature)
    return sig, return_annotation


def setup(app):
    app.connect("builder-inited", doxygen_build)
    app.connect("builder-inited", render_doxygen_pages)
    app.connect("autodoc-process-docstring", _process_pybind_docstrings)
    app.connect("autodoc-process-signature", _process_pybind_signature)



import inspect
import os
import platform
import re
import subprocess
import sys

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install
from setuptools.command.test import test
from distutils.version import LooseVersion


REQUIRED_NUMPY_VERSION = "numpy>=1.12.0"
MIN_CPU_CORES = 2


class lazy_dict(dict):
    """When the value associated to a key is a function, then returns
    the function call instead of the function.
    """

    def __getitem__(self, item):
        value = dict.__getitem__(self, item)
        if inspect.isfunction(value):
            return value()
        return value


def get_sphinx_command():
    """Lazy load of Sphinx distutils command class
    """
    from sphinx.setup_command import BuildDoc

    return BuildDoc


def get_cpu_count():
    try:
        return len(os.sched_getaffinity(0))  # linux only
    except:
        pass

    try:
        return os.cpu_count()  # python 3.4+
    except:
        return 1  # default


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext, object):
    user_options = build_ext.user_options + [
        ('target=', None, "specify the CMake target to build")
    ]

    def initialize_options(self):
        self.target = "sonata_python"
        super(CMakeBuild, self).initialize_options()

    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r"version\s*([\d.]+)", out.decode()).group(1)
            )
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DSONATA_TESTS={}".format(os.environ.get("SONATA_TESTS", "OFF")),
            "-DEXTLIB_FROM_SUBMODULES=ON",
            "-DSONATA_PYTHON=ON",
            "-DSONATA_VERSION=" + self.distribution.get_version(),
            "-DCMAKE_BUILD_TYPE=",
            "-DSONATA_CXX_WARNINGS=OFF",
            '-DPYTHON_EXECUTABLE=' + sys.executable
        ]

        optimize = "OFF" if self.debug else "ON"
        build_args = ["--config", optimize, "--target", self.target]

        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(
                    optimize.upper(), extdir
                )
            ]
            if sys.maxsize > 2 ** 32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            build_args += ["--", "-j{}".format(max(MIN_CPU_CORES, get_cpu_count()))]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


class PkgTest(test):
    """Custom disutils command that acts like as a replacement
    for the "test" command.
    """

    new_commands = [('test_ext', lambda self: True), ('test_doc', lambda self: True)]
    sub_commands = test.sub_commands + new_commands

    def run(self):
        super(PkgTest, self).run()
        self.run_command('test_ext')
        self.run_command('test_doc')


install_requires = [
    REQUIRED_NUMPY_VERSION,
]

setup_requires = [
    "setuptools_scm",
]

# HACK: `setup_requires` uses `easy_install` to do the 'pre-installation' of
# numpy, it doesn't realize it has to install numpy version that supports py2.
if sys.version_info[0] == 3:
    setup_requires.append(REQUIRED_NUMPY_VERSION)

with open('README.rst') as f:
    README = f.read()

setup(
    name="libsonata",
    description='SONATA files reader',
    author="Blue Brain Project, EPFL",
    long_description=README,
    long_description_content_type='text/x-rst',
    license="LGPLv3",
    url='https://github.com/BlueBrain/libsonata',
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
    ext_modules=[CMakeExtension("libsonata._libsonata")],
    cmdclass=lazy_dict(
        build_ext=CMakeBuild,
        test_ext=CMakeBuild,
        test=PkgTest,
        test_doc=get_sphinx_command,
    ),
    zip_safe=False,
    setup_requires=setup_requires,
    install_requires=install_requires,
    extras_require={
        'docs': ['sphinx-bluebrain-theme'],
    },
    use_scm_version={"local_scheme": "no-local-version",
                     },
    package_dir={"": "python"},
    packages=['libsonata',
              ],
)



#!/usr/bin/env python

import sys
import numpy as np
import h5py

def get_vlen_str_type():
    if sys.version_info[0] == 2:
        vlen_str_type = unicode
    elif sys.version_info[0] == 3:
        vlen_str_type = str
    else:
        raise 'Only python2/3 supported'
    return vlen_str_type

def write_population_one_group(pop, prefix):
    string_dtype = h5py.special_dtype(vlen=get_vlen_str_type())

    pop.create_dataset('%s_group_id' % prefix, data=np.zeros(6), dtype=np.uint8)
    pop.create_dataset('%s_group_index' % prefix, data=np.arange(6), dtype=np.uint16)
    pop.create_dataset('%s_type_id' % prefix, data=np.full(6, -1), dtype=np.int32)

    attrs = pop.create_group('0')
    attrs.create_dataset('attr-X', data=np.arange(11., 17.), dtype=np.float64)
    attrs.create_dataset('attr-Y', data=np.arange(21, 27), dtype=np.int64)
    attrs.create_dataset('attr-Z', data=[(2 * x).encode('utf-8') for x in 'abcdef'],
                         dtype=string_dtype)

    # All supported dtypes
    attrs.create_dataset('A-double', dtype=np.float64)
    attrs.create_dataset('A-float', dtype=np.float32)
    attrs.create_dataset('A-int64', dtype=np.int64)
    attrs.create_dataset('A-int32', dtype=np.int32)
    attrs.create_dataset('A-int16', dtype=np.int16)
    attrs.create_dataset('A-int8', dtype=np.int8)
    attrs.create_dataset('A-uint64', dtype=np.uint64)
    attrs.create_dataset('A-uint32', dtype=np.uint32)
    attrs.create_dataset('A-uint16', dtype=np.uint16)
    attrs.create_dataset('A-uint8', dtype=np.uint8)
    attrs.create_dataset('A-string', dtype=string_dtype)

    # Unsupported dtype
    enum_dtype = h5py.special_dtype(enum=('i', {"RED": 0, "GREEN": 1}))
    attrs.create_dataset('A-enum', data=[0, 1], dtype=enum_dtype)

    attrs.create_dataset('E-mapping-good', data=[2, 1, 2, 0, 2, 2], dtype=np.int64)
    attrs.create_dataset('E-mapping-bad', data=[2, 3, 1, 0, -1, 2], dtype=np.int64)

    mapping = [s.encode('utf-8') for s in ("A", "B", "C")]
    library = attrs.create_group('@library')
    library.create_dataset('E-mapping-good', data=mapping, dtype=string_dtype)
    library.create_dataset('E-mapping-bad', data=mapping, dtype=string_dtype)
    library.create_dataset('E-mapping-ugly', data=mapping, dtype=string_dtype)

    dparams = attrs.create_group('dynamics_params')
    dparams.create_dataset('dparam-X', data=np.arange(1011., 1017.), dtype=np.float64)
    dparams.create_dataset('dparam-Y', data=np.arange(1021, 1027), dtype=np.int64)
    dparams.create_dataset('dparam-Z', data=[('d-' + 2 * x).encode('utf-8') for x in 'abcdef'],
                           dtype=string_dtype)


def write_population_two_groups(pop):
    pop.create_group('0')
    pop.create_group('1')


def write_nodes(filepath):
    with h5py.File(filepath, 'w') as h5f:
        root = h5f.create_group('nodes')
        write_population_one_group(
            root.create_group('nodes-A'),
            prefix='node'
        )
        write_population_two_groups(
            root.create_group('nodes-B')
        )


def group_ranges(values):
    result = []
    a, b = 0, 0
    for v in values:
        if v == b:
            b += 1
        else:
            if a < b:
                result.append((a, b))
            a, b = v, v + 1
    if a < b:
        result.append((a, b))
    return result


def write_edge_index_one_side(ids, N, out):
    index1, index2 = [], []
    for _id in range(N):
        ranges = group_ranges(np.where(ids == _id)[0])
        index1.append((len(index2), len(index2) + len(ranges)))
        index2.extend(ranges)
    out.create_dataset('node_id_to_ranges', data=index1, dtype=np.uint64)
    out.create_dataset('range_to_edge_id', data=index2, dtype=np.uint64)


def write_edge_index(pop, source_N, target_N):
    index_group = pop.create_group('indices')
    write_edge_index_one_side(
        pop['source_node_id'][:],
        source_N,
        index_group.create_group('source_to_target')
    )
    write_edge_index_one_side(
        pop['target_node_id'][:],
        target_N,
        index_group.create_group('target_to_source')
    )


def write_edges(filepath):
    with h5py.File(filepath, 'w') as h5f:
        root = h5f.create_group('edges')
        edges_AB = root.create_group('edges-AB')
        write_population_one_group(edges_AB, prefix='edge')
        edges_AB.create_dataset(
            'source_node_id', data=[1, 1, 2, 2, 3, 3], dtype=np.uint64
        ).attrs['node_population'] = u'nodes-A'
        edges_AB.create_dataset(
            'target_node_id', data=[1, 2, 1, 1, 0, 2], dtype=np.uint64
        ).attrs['node_population'] = u'nodes-B'
        write_edge_index(edges_AB, 4, 4)
        write_population_two_groups(
            root.create_group('edges-AC')
        )


def write_soma_report(filepath):
    population_names = ['All', 'soma1', 'soma2']
    node_ids = np.arange(1, 21)
    index_pointers = np.arange(0, 21)
    element_ids = np.zeros(20)
    times = (0.0, 1.0, 0.1)
    data = [node_ids + j*0.1 for j in range(10)]
    string_dtype = h5py.special_dtype(vlen=get_vlen_str_type())
    with h5py.File(filepath, 'w') as h5f:
        root = h5f.create_group('report')
        gpop_all = h5f.create_group('/report/' + population_names[0])
        ddata = gpop_all.create_dataset('data', data=data, dtype=np.float32)
        ddata.attrs.create('units', data="mV", dtype=string_dtype)
        gmapping = h5f.create_group('/report/' + population_names[0] + '/mapping')

        dnodes = gmapping.create_dataset('node_ids', data=node_ids, dtype=np.uint64)
        dnodes.attrs.create('sorted', data=True, dtype=np.uint8)
        gmapping.create_dataset('index_pointers', data=index_pointers, dtype=np.uint64)
        gmapping.create_dataset('element_ids', data=element_ids, dtype=np.uint32)
        dtimes = gmapping.create_dataset('time', data=times, dtype=np.double)
        dtimes.attrs.create('units', data="ms", dtype=string_dtype)

        gpop_soma1 = h5f.create_group('/report/' + population_names[1])
        gpop_soma2 = h5f.create_group('/report/' + population_names[2])


def write_element_report(filepath):
    population_names = ['All', 'element1', 'element42']
    node_ids = np.arange(1, 21)
    index_pointers = np.arange(0, 105, 5)
    repeated = np.arange(50)
    element_ids = np.repeat(repeated, 2)
    times = (0.0, 4.0, 0.2)
    string_dtype = h5py.special_dtype(vlen=get_vlen_str_type())
    with h5py.File(filepath, 'w') as h5f:
        root = h5f.create_group('report')
        gpop_all = h5f.create_group('/report/' + population_names[0])
        d1 = np.arange(0.0, 200, 0.1).reshape(20, -1)
        ddata = gpop_all.create_dataset('data', data=d1, dtype=np.float32)
        ddata.attrs.create('units', data="mV", dtype=string_dtype)
        gmapping = h5f.create_group('/report/' + population_names[0] + '/mapping')

        dnodes = gmapping.create_dataset('node_ids', data=node_ids, dtype=np.uint64)
        dnodes.attrs.create('sorted', data=True, dtype=np.uint8)
        gmapping.create_dataset('index_pointers', data=index_pointers, dtype=np.uint64)
        gmapping.create_dataset('element_ids', data=element_ids, dtype=np.uint32)
        dtimes = gmapping.create_dataset('time', data=times, dtype=np.double)
        dtimes.attrs.create('units', data="ms", dtype=string_dtype)

        gpop_soma1 = h5f.create_group('/report/' + population_names[1])
        gpop_soma2 = h5f.create_group('/report/' + population_names[2])

def write_spikes(filepath):
    population_names = ['All', 'spikes1', 'spikes2', 'empty']
    timestamps_base = (0.3, 0.1, 0.2, 1.3, 0.7)
    node_ids_base = (3, 5, 2, 3, 2)

    sorting_type = h5py.enum_dtype({"none": 0, "by_id": 1, "by_time": 2})
    string_dtype = h5py.special_dtype(vlen=get_vlen_str_type())

    with h5py.File(filepath, 'w') as h5f:
        root = h5f.create_group('spikes')
        gpop_all = h5f.create_group('/spikes/' + population_names[0])
        gpop_all.attrs.create('sorting', data=2, dtype=sorting_type)
        timestamps, node_ids = zip(*sorted(zip(timestamps_base, node_ids_base)))
        set = gpop_all.create_dataset('timestamps', data=timestamps, dtype=np.double)
        gpop_all.create_dataset('node_ids', data=node_ids, dtype=np.uint64)

        gpop_spikes1 = h5f.create_group('/spikes/' + population_names[1])
        gpop_spikes1.attrs.create('sorting', data=1, dtype=sorting_type)
        node_ids, timestamps = zip(*sorted(zip(node_ids_base, timestamps_base)))
        gpop_spikes1.create_dataset('timestamps', data=timestamps, dtype=np.double)
        gpop_spikes1.create_dataset('node_ids', data=node_ids, dtype=np.uint64)

        gpop_spikes2 = h5f.create_group('/spikes/' + population_names[2])
        gpop_spikes2.attrs.create('sorting', data=0, dtype=sorting_type)
        dtimestamps = gpop_spikes2.create_dataset('timestamps', data=timestamps_base, dtype=np.double)
        dtimestamps.attrs.create('units', data="ms", dtype=string_dtype)
        gpop_spikes2.create_dataset('node_ids', data=node_ids_base, dtype=np.uint64)

        gpop_empty = h5f.create_group('/spikes/' + population_names[3])
        gpop_empty.attrs.create('sorting', data=1, dtype=sorting_type)
        dtimestamps = gpop_empty.create_dataset('timestamps', data=[], dtype=np.double)
        gpop_empty.create_dataset('node_ids', data=[], dtype=np.uint64)

if __name__ == '__main__':
    write_nodes('nodes1.h5')
    write_edges('edges1.h5')
    write_spikes('spikes.h5')
    write_soma_report('somas.h5')
    write_element_report('elements.h5')
